@@ -1,7 +1,6 @@
 import logging
 import sys
 from collections import defaultdict
-from functools import partial
 from itertools import chain, islice, repeat
 from logging import StreamHandler
 from types import SimpleNamespace
@@ -31,7 +30,7 @@ def buffered_hook_manager(header_template, get_pos, cond_refresh, term):
 
     def flush(stream):
         if buffers[stream]:
-            write(stream, '\n')
+            write(stream, '\n')  # when the current index is about to change, send a newline.
             stream.flush()
 
     def write(stream, part):
@@ -40,41 +39,62 @@ def buffered_hook_manager(header_template, get_pos, cond_refresh, term):
 
         buffer = buffers[stream]
         if part != '\n':
+            if part.startswith('\x1b'):  # if the part starts with ESC, just send it.
+                stream.write(part)
+                return
             # this will generate a sequence of lines interspersed with None, which will later
             # be rendered as the indent filler to align additional lines under the same header.
             gen = chain.from_iterable(zip(repeat(None), part.splitlines(True)))
             buffer.extend(islice(gen, 1, None))
         else:
             header = get_header()
+            spacer = ' ' * len(header)
+            nested = ''.join(line or spacer for line in buffer)
+            text = f'{header}{nested.rstrip()}\n'
             with cond_refresh:
-                spacer = ' ' * len(header)
-                nested = ''.join(line or spacer for line in buffer)
-                text = f'{header}{nested.rstrip()}\n'
                 if stream in base:  # pragma: no cover
-                    # use the current terminal abstraction for preparing the screen.
                     term.clear_line()
-                # handle all streams, both screen and logging.
+                    term.clear_end_screen()
                 stream.write(text)
                 stream.flush()
                 cond_refresh.notify()
                 buffer[:] = []
 
+    # better hook impl, which works even when nested, since __hash__ will be forwarded.
+    class Hook:
+        def __init__(self, stream):
+            self.__stream = stream
+
+        def write(self, part):
+            return write(self.__stream, part)
+
+        def flush(self):
+            return flush(self.__stream)
+
+        def __getattr__(self, item):
+            return getattr(self.__stream, item)
+
     def get_hook_for(handler):
         if handler.stream:  # supports FileHandlers with delay=true.
             handler.stream.flush()
-        return SimpleNamespace(write=partial(write, handler.stream),
-                               flush=partial(flush, handler.stream),
-                               isatty=sys.stdout.isatty)
+        return Hook(handler.stream)
 
     def install():
         def get_all_loggers():
             yield logging.root
             yield from (logging.getLogger(name) for name in logging.root.manager.loggerDict)
 
+        def set_hook(h):
+            try:
+                return h.setStream(get_hook_for(h))
+            except AttributeError:
+                pass  # ignores errors and return None
+
+        # account for reused handlers within loggers.
+        handlers = set(h for logger in get_all_loggers()
+                       for h in logger.handlers if isinstance(h, StreamHandler))
         # modify all stream handlers, including their subclasses.
-        before_handlers.update({h: _set_stream(h, get_hook_for(h))  # noqa
-                                for logger in get_all_loggers()
-                                for h in logger.handlers if isinstance(h, StreamHandler)})
+        before_handlers.update({h: set_hook(h) for h in handlers})  # there can be Nones now.
         sys.stdout, sys.stderr = (get_hook_for(SimpleNamespace(stream=x)) for x in base)
 
     def uninstall():
@@ -82,14 +102,14 @@ def buffered_hook_manager(header_template, get_pos, cond_refresh, term):
         buffers.clear()
         sys.stdout, sys.stderr = base
 
-        [_set_stream(handler, original_stream)
-         for handler, original_stream in before_handlers.items()]
+        [handler.setStream(original) for handler, original in before_handlers.items() if original]
         before_handlers.clear()
 
-        # does the number of logging handlers changed??
+        # did the number of logging handlers change??
         # if yes, it probably means logging was initialized within alive_bar context,
         # and thus there can be an instrumented stdout or stderr within handlers,
         # which causes a TypeError: unhashable type: 'types.SimpleNamespace'...
+        # or simply a logger **reuses** a handler...
 
     # internal data.
     buffers = defaultdict(list)
@@ -127,19 +147,3 @@ def gen_header(header_template, get_pos):  # pragma: no cover
 
 def null_header():  # pragma: no cover
     return ''
-
-
-if sys.version_info >= (3, 7):  # pragma: no cover
-    def _set_stream(handler, stream):
-        return handler.setStream(stream)
-else:  # pragma: no cover
-    def _set_stream(handler, stream):
-        # from python 3.7 implementation.
-        result = handler.stream
-        handler.acquire()
-        try:
-            handler.flush()
-            handler.stream = stream
-        finally:
-            handler.release()
-        return result
